/* -*- Mode: C; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*- */
/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright 2013 Red Hat, Inc.
 */

#include "config.h"

#include <glib/gi18n-lib.h>

#include "nmtui.h"
#include "nmt-connection-list.h"
#include "nmt-connection-utils.h"
#include "nmt-editor.h"

#include "nm-editor-utils.h"

G_DEFINE_TYPE (NmtConnectionList, nmt_connection_list, NMT_TYPE_NEWT_GRID)

#define NMT_CONNECTION_LIST_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), NMT_TYPE_CONNECTION_LIST, NmtConnectionListPrivate))

typedef struct {
	GSList *connections;

	gboolean grouped;
	NmtConnectionListFilter connection_filter;
	gpointer connection_filter_data;

	NmtNewtListbox *listbox;
	NmtNewtGrid *buttons;

	NmtNewtWidget *add;
	NmtNewtWidget *edit;
	NmtNewtWidget *delete;
	NmtNewtWidget *extra;
} NmtConnectionListPrivate;

enum {
	PROP_0,

	PROP_GROUPED,
	PROP_CONNECTION_FILTER,
	PROP_CONNECTION_FILTER_DATA,
	PROP_EXTRA_WIDGET,
	PROP_CONNECTIONS,
	PROP_NUM_CONNECTIONS,

	LAST_PROP
};

enum {
	ADD_CONNECTION,
	EDIT_CONNECTION,
	REMOVE_CONNECTION,

	LAST_SIGNAL
};

static guint signals[LAST_SIGNAL] = { 0 };

static void add_clicked (NmtNewtButton *button, gpointer list);
static void edit_clicked (NmtNewtButton *button, gpointer list);
static void delete_clicked (NmtNewtButton *button, gpointer list);
static void listbox_activated (NmtNewtListbox *listbox, gpointer list);

static void
nmt_connection_list_init (NmtConnectionList *list)
{
	NmtConnectionListPrivate *priv = NMT_CONNECTION_LIST_GET_PRIVATE (list);
	NmtNewtWidget *listbox, *buttons;
	NmtNewtGrid *grid = NMT_NEWT_GRID (list);

	listbox = g_object_new (NMT_TYPE_NEWT_LISTBOX,
	                        "flags", NMT_NEWT_LISTBOX_SCROLL | NMT_NEWT_LISTBOX_BORDER,
	                        "skip-null-keys", TRUE,
	                        NULL);
	priv->listbox = NMT_NEWT_LISTBOX (listbox);
	nmt_newt_grid_add (grid, listbox, 0, 0);
	nmt_newt_grid_set_flags (grid, listbox,
	                         NMT_NEWT_GRID_FILL_X | NMT_NEWT_GRID_FILL_Y |
	                         NMT_NEWT_GRID_EXPAND_X | NMT_NEWT_GRID_EXPAND_Y);
	g_signal_connect (priv->listbox, "activated", G_CALLBACK (listbox_activated), list);

	buttons = nmt_newt_grid_new ();
	priv->buttons = NMT_NEWT_GRID (buttons);
	nmt_newt_grid_add (grid, buttons, 1, 0);
	nmt_newt_grid_set_padding (grid, buttons, 1, 0, 0, 0);
	nmt_newt_grid_set_flags (grid, buttons,
	                         NMT_NEWT_GRID_FILL_X | NMT_NEWT_GRID_FILL_Y |
	                         NMT_NEWT_GRID_EXPAND_Y);

	priv->add = nmt_newt_button_new (_("Add"));
	nmt_newt_grid_add (priv->buttons, priv->add, 0, 0);
	nmt_newt_grid_set_padding (priv->buttons, priv->add, 0, 1, 0, 0);
	g_signal_connect (priv->add, "clicked", G_CALLBACK (add_clicked), list);

	priv->edit = nmt_newt_button_new (_("Edit..."));
	nmt_newt_grid_add (priv->buttons, priv->edit, 0, 1);
	nmt_newt_grid_set_padding (priv->buttons, priv->edit, 0, 1, 0, 0);
	g_signal_connect (priv->edit, "clicked", G_CALLBACK (edit_clicked), list);

	priv->delete = nmt_newt_button_new (_("Delete"));
	nmt_newt_grid_add (priv->buttons, priv->delete, 0, 2);
	nmt_newt_grid_set_padding (priv->buttons, priv->delete, 0, 1, 0, 1);
	g_signal_connect (priv->delete, "clicked", G_CALLBACK (delete_clicked), list);
}

static int
sort_by_timestamp (gconstpointer  a,
                   gconstpointer  b)
{
	NMSettingConnection *s_con_a, *s_con_b;
	guint64 time_a, time_b;

	s_con_a = nm_connection_get_setting_connection ((NMConnection *) a);
	s_con_b = nm_connection_get_setting_connection ((NMConnection *) b);

	time_a = nm_setting_connection_get_timestamp (s_con_a);
	time_b = nm_setting_connection_get_timestamp (s_con_b);

	return (int) (time_b - time_a);
}

static void
nmt_connection_list_rebuild (NmtConnectionList *list)
{
	NmtConnectionListPrivate *priv = NMT_CONNECTION_LIST_GET_PRIVATE (list);
	NmtNewtListbox *listbox;
	GSList *iter, *next;
	gboolean did_header = FALSE, did_vpn = FALSE;
	NMEditorConnectionTypeData **types;
	int i;

	g_slist_free_full (priv->connections, g_object_unref);
	priv->connections = nm_remote_settings_list_connections (nm_settings);
	for (iter = priv->connections; iter; iter = next) {
		next = iter->next;

		if (priv->connection_filter) {
			if (!priv->connection_filter (list, iter->data, priv->connection_filter_data))
				priv->connections = g_slist_delete_link (priv->connections, iter);
			else
				g_object_ref (iter->data);
		} else
			g_object_ref (iter->data);
	}
	priv->connections = g_slist_sort (priv->connections, sort_by_timestamp);
	g_object_notify (G_OBJECT (list), "connections");
	g_object_notify (G_OBJECT (list), "num-connections");

	nmt_newt_component_set_sensitive (NMT_NEWT_COMPONENT (priv->edit),
	                                  priv->connections != NULL);
	nmt_newt_component_set_sensitive (NMT_NEWT_COMPONENT (priv->delete),
	                                  priv->connections != NULL);

	listbox = NMT_NEWT_LISTBOX (priv->listbox);
	nmt_newt_listbox_clear (listbox);

	if (!priv->grouped) {
		/* Just add the connections in order */
		for (iter = priv->connections; iter; iter = iter->next) {
			NMConnection *conn = iter->data;
			nmt_newt_listbox_append (listbox, nm_connection_get_id (conn), conn);
		}
		return;
	}

	types = nm_editor_utils_get_connection_type_list ();
	for (i = 0; types[i]; i++) {
		if (types[i]->setting_type == NM_TYPE_SETTING_VPN) {
			if (did_vpn)
				continue;
			did_vpn = TRUE;
		}

		did_header = FALSE;

		for (iter = priv->connections; iter; iter = iter->next) {
			NMConnection *conn = iter->data;
			NMSetting *setting;
			char *indented;

			setting = nm_connection_get_setting (conn, types[i]->setting_type);
			if (!setting)
				continue;
			if (!nm_connection_is_type (conn, nm_setting_get_name (setting)))
				continue;

			if (!did_header) {
				nmt_newt_listbox_append (listbox, types[i]->name, NULL);
				did_header = TRUE;
			}

			indented = g_strdup_printf ("  %s", nm_connection_get_id (conn));
			nmt_newt_listbox_append (listbox, indented, conn);
			g_free (indented);
		}
	}
}

static void
nmt_connection_list_constructed (GObject *object)
{
	NmtConnectionList *list = NMT_CONNECTION_LIST (object);
	NmtConnectionListPrivate *priv = NMT_CONNECTION_LIST_GET_PRIVATE (list);

	if (priv->extra) {
		nmt_newt_grid_add (priv->buttons, priv->extra, 0, 3);
		nmt_newt_grid_set_padding (priv->buttons, priv->extra, 0, 3, 0, 1);
		nmt_newt_grid_set_flags (priv->buttons, priv->extra,
		                         NMT_NEWT_GRID_FILL_X |
		                         NMT_NEWT_GRID_EXPAND_Y | NMT_NEWT_GRID_ANCHOR_BOTTOM);
	}

	nmt_connection_list_rebuild (list);

	G_OBJECT_CLASS (nmt_connection_list_parent_class)->constructed (object);
}

static void
add_clicked (NmtNewtButton *button, gpointer list)
{
	g_signal_emit (list, signals[ADD_CONNECTION], 0);
	nmt_connection_list_rebuild (list);
}

static void
edit_clicked (NmtNewtButton *button, gpointer list)
{
	NmtConnectionListPrivate *priv = NMT_CONNECTION_LIST_GET_PRIVATE (list);
	NMConnection *connection;

	connection = nmt_newt_listbox_get_active_key (priv->listbox);
	g_return_if_fail (connection != NULL);

	g_signal_emit (list, signals[EDIT_CONNECTION], 0, connection);
	nmt_connection_list_rebuild (list);
}

static void
delete_clicked (NmtNewtButton *button, gpointer list)
{
	NmtConnectionListPrivate *priv = NMT_CONNECTION_LIST_GET_PRIVATE (list);
	NMRemoteConnection *connection;

	connection = nmt_newt_listbox_get_active_key (priv->listbox);
	g_return_if_fail (connection != NULL);

	g_signal_emit (list, signals[REMOVE_CONNECTION], 0, connection);
	nmt_connection_list_rebuild (list);
}

static void
listbox_activated (NmtNewtListbox *listbox, gpointer list)
{
	NmtConnectionListPrivate *priv = NMT_CONNECTION_LIST_GET_PRIVATE (list);

	edit_clicked (NMT_NEWT_BUTTON (priv->edit), list);
}

static void
nmt_connection_list_finalize (GObject *object)
{
	NmtConnectionListPrivate *priv = NMT_CONNECTION_LIST_GET_PRIVATE (object);

	g_slist_free_full (priv->connections, g_object_unref);
	g_clear_object (&priv->extra);

	G_OBJECT_CLASS (nmt_connection_list_parent_class)->finalize (object);
}

static void
nmt_connection_list_set_property (GObject      *object,
                                  guint         prop_id,
                                  const GValue *value,
                                  GParamSpec   *pspec)
{
	NmtConnectionListPrivate *priv = NMT_CONNECTION_LIST_GET_PRIVATE (object);

	switch (prop_id) {
	case PROP_GROUPED:
		priv->grouped = g_value_get_boolean (value);
		break;
	case PROP_CONNECTION_FILTER:
		priv->connection_filter = g_value_get_pointer (value);
		break;
	case PROP_CONNECTION_FILTER_DATA:
		priv->connection_filter_data = g_value_get_pointer (value);
		break;
	case PROP_EXTRA_WIDGET:
		priv->extra = g_value_get_object (value);
		if (priv->extra)
			g_object_ref_sink (priv->extra);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

static void
nmt_connection_list_get_property (GObject    *object,
                                  guint       prop_id,
                                  GValue     *value,
                                  GParamSpec *pspec)
{
	NmtConnectionListPrivate *priv = NMT_CONNECTION_LIST_GET_PRIVATE (object);
	GPtrArray *connections;
	GSList *iter;

	switch (prop_id) {
	case PROP_GROUPED:
		g_value_set_boolean (value, priv->grouped);
		break;
	case PROP_CONNECTION_FILTER:
		g_value_set_pointer (value, priv->connection_filter);
		break;
	case PROP_CONNECTION_FILTER_DATA:
		g_value_set_pointer (value, priv->connection_filter_data);
		break;
	case PROP_EXTRA_WIDGET:
		g_value_set_object (value, priv->extra);
		break;
	case PROP_CONNECTIONS:
		connections = g_ptr_array_new_with_free_func (g_object_unref);
		for (iter = priv->connections; iter; iter = iter->next)
			g_ptr_array_add (connections, g_object_ref (iter->data));
		g_value_take_boxed (value, connections);
		break;
	case PROP_NUM_CONNECTIONS:
		g_value_set_int (value, g_slist_length (priv->connections));
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

static void
nmt_connection_list_class_init (NmtConnectionListClass *list_class)
{
	GObjectClass *object_class = G_OBJECT_CLASS (list_class);

	g_type_class_add_private (list_class, sizeof (NmtConnectionListPrivate));

	/* virtual methods */
	object_class->constructed  = nmt_connection_list_constructed;
	object_class->set_property = nmt_connection_list_set_property;
	object_class->get_property = nmt_connection_list_get_property;
	object_class->finalize     = nmt_connection_list_finalize;

	/* signals */
	signals[ADD_CONNECTION] =
		g_signal_new ("add-connection",
		              G_OBJECT_CLASS_TYPE (object_class),
		              G_SIGNAL_RUN_LAST,
		              G_STRUCT_OFFSET (NmtConnectionListClass, add_connection),
		              NULL, NULL, NULL,
		              G_TYPE_NONE, 0);
	signals[EDIT_CONNECTION] =
		g_signal_new ("edit-connection",
		              G_OBJECT_CLASS_TYPE (object_class),
		              G_SIGNAL_RUN_LAST,
		              G_STRUCT_OFFSET (NmtConnectionListClass, edit_connection),
		              NULL, NULL, NULL,
		              G_TYPE_NONE, 1,
		              NM_TYPE_CONNECTION);
	signals[REMOVE_CONNECTION] =
		g_signal_new ("remove-connection",
		              G_OBJECT_CLASS_TYPE (object_class),
		              G_SIGNAL_RUN_LAST,
		              G_STRUCT_OFFSET (NmtConnectionListClass, remove_connection),
		              NULL, NULL, NULL,
		              G_TYPE_NONE, 1,
		              NM_TYPE_CONNECTION);

	/* properties */
	g_object_class_install_property (object_class, PROP_GROUPED,
	                                 g_param_spec_boolean ("grouped", "", "",
	                                                       TRUE,
	                                                       G_PARAM_READWRITE |
	                                                       G_PARAM_CONSTRUCT_ONLY |
	                                                       G_PARAM_STATIC_STRINGS));
	g_object_class_install_property (object_class, PROP_CONNECTION_FILTER,
	                                 g_param_spec_pointer ("connection-filter", "", "",
	                                                       G_PARAM_READWRITE |
	                                                       G_PARAM_CONSTRUCT_ONLY |
	                                                       G_PARAM_STATIC_STRINGS));
	g_object_class_install_property (object_class, PROP_CONNECTION_FILTER_DATA,
	                                 g_param_spec_pointer ("connection-filter-data", "", "",
	                                                       G_PARAM_READWRITE |
	                                                       G_PARAM_CONSTRUCT_ONLY |
	                                                       G_PARAM_STATIC_STRINGS));
	g_object_class_install_property (object_class, PROP_EXTRA_WIDGET,
	                                 g_param_spec_object ("extra-widget", "", "",
	                                                      NMT_TYPE_NEWT_WIDGET,
	                                                      G_PARAM_READWRITE |
	                                                      G_PARAM_CONSTRUCT_ONLY |
	                                                      G_PARAM_STATIC_STRINGS));
	g_object_class_install_property (object_class, PROP_CONNECTIONS,
	                                 g_param_spec_boxed ("connections", "", "",
	                                                     G_TYPE_PTR_ARRAY,
	                                                     G_PARAM_READABLE |
	                                                     G_PARAM_STATIC_STRINGS));
	g_object_class_install_property (object_class, PROP_NUM_CONNECTIONS,
	                                 g_param_spec_int ("num-connections", "", "",
	                                                   0, G_MAXINT, 0,
	                                                   G_PARAM_READABLE |
	                                                   G_PARAM_STATIC_STRINGS));
}
