/* -*- Mode: C; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*- */
/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright 2013 Red Hat, Inc.
 */

#include "config.h"

#include <string.h>

#include "nmt-newt-entry.h"
#include "nmt-newt-form.h"
#include "nmt-newt-hacks.h"
#include "nmt-newt-utils.h"

G_DEFINE_TYPE (NmtNewtEntry, nmt_newt_entry, NMT_TYPE_NEWT_COMPONENT)

#define NMT_NEWT_ENTRY_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), NMT_TYPE_NEWT_ENTRY, NmtNewtEntryPrivate))

typedef struct {
	int width;
	NmtNewtEntryFlags flags;
	char *text;
	int last_cursor_pos;

	NmtNewtEntryFilter filter;
	gpointer filter_data;

	NmtNewtEntryValidator validator;
	gpointer validator_data;
} NmtNewtEntryPrivate;

enum {
	PROP_0,
	PROP_TEXT,
	PROP_WIDTH,
	PROP_FLAGS,
	PROP_PASSWORD,

	LAST_PROP
};

NmtNewtWidget *
nmt_newt_entry_new (int               width,
                    NmtNewtEntryFlags flags)
{
	return g_object_new (NMT_TYPE_NEWT_ENTRY,
	                     "width", width,
	                     "flags", flags,
	                     NULL);
}

void
nmt_newt_entry_set_filter (NmtNewtEntry       *entry,
                           NmtNewtEntryFilter  filter,
                           gpointer            user_data)
{
	NmtNewtEntryPrivate *priv = NMT_NEWT_ENTRY_GET_PRIVATE (entry);

	priv->filter = filter;
	priv->filter_data = user_data;
}

static void
nmt_newt_entry_check_valid (NmtNewtEntry *entry)
{
	NmtNewtEntryPrivate *priv = NMT_NEWT_ENTRY_GET_PRIVATE (entry);
	gboolean valid;

	if (   (priv->flags & NMT_NEWT_ENTRY_NONEMPTY)
	    && *priv->text == '\0')
		valid = FALSE;
	else if (priv->validator)
		valid = !!priv->validator (entry, priv->text, priv->validator_data);
	else
		valid = TRUE;

	nmt_newt_widget_set_valid (NMT_NEWT_WIDGET (entry), valid);
}

void
nmt_newt_entry_set_validator (NmtNewtEntry          *entry,
                              NmtNewtEntryValidator  validator,
                              gpointer               user_data)
{
	NmtNewtEntryPrivate *priv = NMT_NEWT_ENTRY_GET_PRIVATE (entry);

	priv->validator = validator;
	priv->validator_data = user_data;

	nmt_newt_entry_check_valid (entry);
}

static void
nmt_newt_entry_set_text_internal (NmtNewtEntry  *entry,
                                  const char    *text,
                                  newtComponent  co)
{
	NmtNewtEntryPrivate *priv = NMT_NEWT_ENTRY_GET_PRIVATE (entry);

	if (!text)
		text = "";

	if (!strcmp (priv->text, text))
		return;

	g_free (priv->text);
	priv->text = g_strdup (text);

	if (co) {
		char *text_lc;

		text_lc = priv->text ? nmt_newt_locale_from_utf8 (priv->text) : NULL;
		newtEntrySet (co, text_lc, TRUE);
		g_free (text_lc);
		priv->last_cursor_pos = -1;
	}

	g_object_freeze_notify (G_OBJECT (entry));
	nmt_newt_entry_check_valid (entry);
	g_object_notify (G_OBJECT (entry), "text");
	g_object_thaw_notify (G_OBJECT (entry));
}

void
nmt_newt_entry_set_text (NmtNewtEntry *entry,
                         const char   *text)
{
	newtComponent co;

	co = nmt_newt_component_get_component (NMT_NEWT_COMPONENT (entry));
	nmt_newt_entry_set_text_internal (entry, text, co);
}

const char *
nmt_newt_entry_get_text (NmtNewtEntry *entry)
{
	NmtNewtEntryPrivate *priv = NMT_NEWT_ENTRY_GET_PRIVATE (entry);

	return priv->text;
}

void
nmt_newt_entry_set_width (NmtNewtEntry *entry,
                          int           width)
{
	NmtNewtEntryPrivate *priv = NMT_NEWT_ENTRY_GET_PRIVATE (entry);

	if (priv->width == width)
		return;

	priv->width = width;
	nmt_newt_widget_needs_rebuild (NMT_NEWT_WIDGET (entry));

	g_object_notify (G_OBJECT (entry), "width");
}

int
nmt_newt_entry_get_width (NmtNewtEntry *entry)
{
	NmtNewtEntryPrivate *priv = NMT_NEWT_ENTRY_GET_PRIVATE (entry);

	return priv->width;
}

static void
nmt_newt_entry_init (NmtNewtEntry *entry)
{
	NmtNewtEntryPrivate *priv = NMT_NEWT_ENTRY_GET_PRIVATE (entry);

	priv->text = g_strdup ("");
	priv->last_cursor_pos = -1;
}

static void
nmt_newt_entry_constructed (GObject *object)
{
	nmt_newt_entry_check_valid (NMT_NEWT_ENTRY (object));

	G_OBJECT_CLASS (nmt_newt_entry_parent_class)->constructed (object);
}

static void
nmt_newt_entry_finalize (GObject *object)
{
	NmtNewtEntryPrivate *priv = NMT_NEWT_ENTRY_GET_PRIVATE (object);

	g_free (priv->text);

	G_OBJECT_CLASS (nmt_newt_entry_parent_class)->finalize (object);
}

static int
entry_filter (newtComponent  entry,
              void          *self,
              int            ch,
              int            cursor)
{
	NmtNewtEntryPrivate *priv = NMT_NEWT_ENTRY_GET_PRIVATE (self);

	if (g_ascii_isprint (ch)) {
		if (priv->filter) {
			char *text = nmt_newt_locale_to_utf8 (newtEntryGetValue (entry));

			if (!priv->filter (self, text, ch, cursor, priv->filter_data)) {
				g_free (text);
				return 0;
			}
			g_free (text);
		}
	}

	nmt_newt_form_widget_activated (nmt_newt_widget_get_form (self), self, entry);
	return ch;
}

static guint
convert_flags (NmtNewtEntryFlags flags)
{
	guint newt_flags = 0;

	if (!(flags & NMT_NEWT_ENTRY_NOSCROLL))
		newt_flags |= NEWT_FLAG_SCROLL;
	if (flags & NMT_NEWT_ENTRY_PASSWORD)
		newt_flags |= NEWT_FLAG_PASSWORD;

	return newt_flags;
}

static newtComponent
nmt_newt_entry_build_component (NmtNewtComponent *component,
                                gboolean          sensitive)
{
	NmtNewtEntryPrivate *priv = NMT_NEWT_ENTRY_GET_PRIVATE (component);
	newtComponent co;
	char *text_lc;
	int flags;

	flags = convert_flags (priv->flags);
	if (!sensitive)
		flags |= NEWT_FLAG_DISABLED;

	text_lc = priv->text ? nmt_newt_locale_from_utf8 (priv->text) : NULL;
	co = newtEntry (-1, -1, text_lc, priv->width, NULL, flags);
	g_free (text_lc);

	if (priv->last_cursor_pos != -1)
		newtEntrySetCursorPosition (co, priv->last_cursor_pos);

	newtEntrySetFilter (co, entry_filter, component);
	return co;
}

static void
nmt_newt_entry_activate_component (NmtNewtComponent *component,
                                   newtComponent     co)
{
	NmtNewtEntryPrivate *priv = NMT_NEWT_ENTRY_GET_PRIVATE (component);
	const char *text_lc = newtEntryGetValue (co);
	char *text = nmt_newt_locale_to_utf8 (text_lc);

	priv->last_cursor_pos = newtEntryGetCursorPosition (co);
	nmt_newt_entry_set_text_internal (NMT_NEWT_ENTRY (component), text, NULL);
	g_free (text);
}

static void
nmt_newt_entry_set_property (GObject      *object,
                             guint         prop_id,
                             const GValue *value,
                             GParamSpec   *pspec)
{
	NmtNewtEntry *entry = NMT_NEWT_ENTRY (object);
	NmtNewtEntryPrivate *priv = NMT_NEWT_ENTRY_GET_PRIVATE (entry);

	switch (prop_id) {
	case PROP_TEXT:
		nmt_newt_entry_set_text (entry, g_value_get_string (value));
		break;
	case PROP_WIDTH:
		nmt_newt_entry_set_width (entry, g_value_get_int (value));
		break;
	case PROP_FLAGS:
		priv->flags = g_value_get_uint (value);
		nmt_newt_widget_needs_rebuild (NMT_NEWT_WIDGET (entry));
		break;
	case PROP_PASSWORD:
		if (g_value_get_boolean (value))
			priv->flags |= NMT_NEWT_ENTRY_PASSWORD;
		else
			priv->flags &= ~NMT_NEWT_ENTRY_PASSWORD;
		nmt_newt_widget_needs_rebuild (NMT_NEWT_WIDGET (entry));
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

static void
nmt_newt_entry_get_property (GObject    *object,
                             guint       prop_id,
                             GValue     *value,
                             GParamSpec *pspec)
{
	NmtNewtEntry *entry = NMT_NEWT_ENTRY (object);
	NmtNewtEntryPrivate *priv = NMT_NEWT_ENTRY_GET_PRIVATE (entry);

	switch (prop_id) {
	case PROP_TEXT:
		g_value_set_string (value, nmt_newt_entry_get_text (entry));
		break;
	case PROP_WIDTH:
		g_value_set_int (value, priv->width);
		break;
	case PROP_FLAGS:
		g_value_set_uint (value, priv->flags);
		break;
	case PROP_PASSWORD:
		g_value_set_boolean (value, (priv->flags & NMT_NEWT_ENTRY_PASSWORD) != 0);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

static void
nmt_newt_entry_class_init (NmtNewtEntryClass *entry_class)
{
	GObjectClass *object_class = G_OBJECT_CLASS (entry_class);
	NmtNewtComponentClass *component_class = NMT_NEWT_COMPONENT_CLASS (entry_class);

	g_type_class_add_private (entry_class, sizeof (NmtNewtEntryPrivate));

	/* virtual methods */
	object_class->constructed  = nmt_newt_entry_constructed;
	object_class->set_property = nmt_newt_entry_set_property;
	object_class->get_property = nmt_newt_entry_get_property;
	object_class->finalize     = nmt_newt_entry_finalize;

	component_class->build_component = nmt_newt_entry_build_component;
	component_class->activate_component = nmt_newt_entry_activate_component;

	g_object_class_install_property (object_class, PROP_TEXT,
	                                 g_param_spec_string ("text", "", "",
	                                                      NULL,
	                                                      G_PARAM_READWRITE |
	                                                      G_PARAM_STATIC_STRINGS));
	g_object_class_install_property (object_class, PROP_WIDTH,
	                                 g_param_spec_int ("width", "", "",
	                                                   -1, 80, -1,
	                                                   G_PARAM_READWRITE |
	                                                   G_PARAM_STATIC_STRINGS));
	g_object_class_install_property (object_class, PROP_FLAGS,
	                                 g_param_spec_uint ("flags", "", "",
	                                                    0, 0xFFFF, 0,
	                                                    G_PARAM_READWRITE |
	                                                    G_PARAM_CONSTRUCT_ONLY |
	                                                    G_PARAM_STATIC_STRINGS));
	g_object_class_install_property (object_class, PROP_PASSWORD,
	                                 g_param_spec_boolean ("password", "", "",
	                                                       FALSE,
	                                                       G_PARAM_READWRITE |
	                                                       G_PARAM_STATIC_STRINGS));
}
