/* -*- Mode: C; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*- */
/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright 2013 Red Hat, Inc.
 */

#include "config.h"

#include <string.h>

#include <glib/gi18n-lib.h>

#include "nmt-newt-form.h"
#include "nmt-newt-button.h"
#include "nmt-newt-grid.h"
#include "nmt-newt-utils.h"

G_DEFINE_TYPE (NmtNewtForm, nmt_newt_form, NMT_TYPE_NEWT_CONTAINER)

#define NMT_NEWT_FORM_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), NMT_TYPE_NEWT_FORM, NmtNewtFormPrivate))

typedef struct {
	newtComponent form;
	NmtNewtWidget *content;

	guint x, y, width, height;
	guint padding;
	gboolean fixed_x, fixed_y;
	gboolean fixed_width, fixed_height;
	char *title_lc;

	gboolean running, dirty, escape_exits;
	NmtNewtWidget *focus;
	NmtNewtWidget *activated;
	newtComponent activated_co;
} NmtNewtFormPrivate;

enum {
	PROP_0,
	PROP_TITLE,
	PROP_FULLSCREEN,
	PROP_X,
	PROP_Y,
	PROP_WIDTH,
	PROP_HEIGHT,
	PROP_PADDING,
	PROP_ESCAPE_EXITS,

	LAST_PROP
};

NmtNewtForm *
nmt_newt_form_new (const char *title)
{
	return g_object_new (NMT_TYPE_NEWT_FORM,
	                     "title", title,
	                     NULL);
}

NmtNewtForm *
nmt_newt_form_new_fullscreen (const char *title)
{
	return g_object_new (NMT_TYPE_NEWT_FORM,
	                     "title", title,
	                     "fullscreen", TRUE,
	                     NULL);
}

static void
nmt_newt_form_init (NmtNewtForm *form)
{
	g_object_ref_sink (form);
}

static void
nmt_newt_form_finalize (GObject *object)
{
	NmtNewtFormPrivate *priv = NMT_NEWT_FORM_GET_PRIVATE (object);

	g_free (priv->title_lc);
	g_clear_object (&priv->focus);

	G_OBJECT_CLASS (nmt_newt_form_parent_class)->finalize (object);
}

static void
nmt_newt_form_needs_rebuild (NmtNewtWidget *widget)
{
	NmtNewtFormPrivate *priv = NMT_NEWT_FORM_GET_PRIVATE (widget);

	if (!priv->dirty) {
		priv->dirty = TRUE;
		if (priv->form)
			newtFormSetTimer (priv->form, 1);
	}
}

static void
nmt_newt_form_remove (NmtNewtContainer *container,
                      NmtNewtWidget    *widget)
{
	NmtNewtFormPrivate *priv = NMT_NEWT_FORM_GET_PRIVATE (container);
	NmtNewtContainerClass *parent_class = NMT_NEWT_CONTAINER_CLASS (nmt_newt_form_parent_class);

	g_return_if_fail (widget == priv->content);

	parent_class->remove (container, widget);
	priv->content = NULL;
}

void
nmt_newt_form_set_content (NmtNewtForm      *form,
                           NmtNewtWidget    *content)
{
	NmtNewtFormPrivate *priv = NMT_NEWT_FORM_GET_PRIVATE (form);
	NmtNewtContainerClass *parent_class = NMT_NEWT_CONTAINER_CLASS (nmt_newt_form_parent_class);

	if (priv->content)
		nmt_newt_form_remove (NMT_NEWT_CONTAINER (form), priv->content);

	priv->content = content;

	if (priv->content)
		parent_class->add (NMT_NEWT_CONTAINER (form), content);
}

static newtComponent
nmt_newt_form_build (NmtNewtForm *form)
{
	NmtNewtFormPrivate *priv = NMT_NEWT_FORM_GET_PRIVATE (form);
	int screen_height, screen_width, form_height, form_width;
	newtComponent newt_form, *cos;
	int i;

	priv->dirty = FALSE;
	nmt_newt_widget_realize (priv->content);

	nmt_newt_widget_size_request (priv->content, &form_width, &form_height);
	newtGetScreenSize (&screen_width, &screen_height);

	if (!priv->fixed_width)
		priv->width = MIN (form_width + 2 * priv->padding, screen_width - 2);
	if (!priv->fixed_height)
		priv->height = MIN (form_height + 2 * priv->padding, screen_height - 2);

	if (!priv->fixed_x)
		priv->x = (screen_width - form_width) / 2;
	if (!priv->fixed_y)
		priv->y = (screen_height - form_height) / 2;

	nmt_newt_widget_size_allocate (priv->content,
	                               priv->padding,
	                               priv->padding,
	                               priv->width - 2 * priv->padding,
	                               priv->height - 2 * priv->padding);

	if (priv->height - 2 * priv->padding < form_height) {
		newtComponent scroll_bar =
			newtVerticalScrollbar (priv->width - 1, 0, priv->height,
			                       NEWT_COLORSET_WINDOW,
			                       NEWT_COLORSET_ACTCHECKBOX);

		newt_form = newtForm (scroll_bar, NULL, NEWT_FLAG_NOF12);
		newtFormAddComponent (newt_form, scroll_bar);
		newtFormSetHeight (newt_form, priv->height - 2);
	} else
		newt_form = newtForm (NULL, NULL, NEWT_FLAG_NOF12);

	if (priv->escape_exits)
		newtFormAddHotKey (newt_form, NEWT_KEY_ESCAPE);

	cos = nmt_newt_widget_get_components (priv->content);
	for (i = 0; cos[i]; i++)
		newtFormAddComponent (newt_form, cos[i]);
	g_free (cos);

	return newt_form;
}

static NmtNewtWidget *
nmt_newt_form_real_run (NmtNewtForm *form)
{
	NmtNewtFormPrivate *priv = NMT_NEWT_FORM_GET_PRIVATE (form);
	NmtNewtWidget *focus;
	newtComponent co;
	struct newtExitStruct es;
#ifdef HAVE_NEWTFORMGETSCROLLPOSITION
	int scroll_position = 0;
#endif

	priv->running = TRUE;
	priv->activated = NULL;
	priv->activated_co = NULL;
	while (priv->running) {
		priv->form = nmt_newt_form_build (form);

		if (priv->focus) {
			newtComponent fco;

			fco = nmt_newt_widget_get_focus_component (priv->focus);
			if (fco)
				newtFormSetCurrent (priv->form, fco);
		}
#ifdef HAVE_NEWTFORMGETSCROLLPOSITION
		if (scroll_position)
			newtFormSetScrollPosition (priv->form, scroll_position);
#endif

		newtOpenWindow (priv->x, priv->y, priv->width, priv->height, priv->title_lc);

		while (priv->running && !priv->dirty) {
			newtFormSetTimer (priv->form, 0);
			newtFormRun (priv->form, &es);

			if (   es.reason == NEWT_EXIT_HOTKEY
			    || es.reason == NEWT_EXIT_ERROR) {
				priv->running = FALSE;
				break;
			} else if (es.reason == NEWT_EXIT_COMPONENT)
				co = es.u.co;
			else
				co = NULL;

			if (co) {
				if (NMT_IS_NEWT_CONTAINER (priv->content)) {
					focus = nmt_newt_container_find_widget (NMT_NEWT_CONTAINER (priv->content), co);
					nmt_newt_form_set_focus (form, focus);
				} else
					nmt_newt_form_set_focus (form, priv->content);
				if (priv->focus) {
					if (!nmt_newt_widget_activate (priv->focus, co))
						nmt_newt_form_set_focus (form, NULL);
				}
			} else {
				co = newtFormGetCurrent (priv->form);
				if (NMT_IS_NEWT_CONTAINER (priv->content)) {
					focus = nmt_newt_container_find_widget (NMT_NEWT_CONTAINER (priv->content), co);
					nmt_newt_form_set_focus (form, focus);
				} else
					nmt_newt_form_set_focus (form, priv->content);
			}

			if (priv->activated) {
				nmt_newt_widget_activate (priv->activated, priv->activated_co);
				priv->activated = NULL;
				priv->activated_co = NULL;
			}
		}

#ifdef HAVE_NEWTFORMGETSCROLLPOSITION
		scroll_position = newtFormGetScrollPosition (priv->form);
#endif

		newtFormDestroy (priv->form);
		nmt_newt_widget_unrealize (priv->content);
		priv->form = NULL;
		newtPopWindow ();
	}

	return priv->focus;
}

NmtNewtWidget *
nmt_newt_form_run (NmtNewtForm *form)
{
	return NMT_NEWT_FORM_GET_CLASS (form)->run (form);
}

void
nmt_newt_form_quit (NmtNewtForm *form)
{
	NmtNewtFormPrivate *priv = NMT_NEWT_FORM_GET_PRIVATE (form);

	g_return_if_fail (priv->form != NULL);

	priv->running = FALSE;
	newtFormSetTimer (priv->form, 1);
}

void
nmt_newt_form_set_focus (NmtNewtForm   *form,
                         NmtNewtWidget *widget)
{
	NmtNewtFormPrivate *priv = NMT_NEWT_FORM_GET_PRIVATE (form);

	g_return_if_fail (priv->form != NULL);

	if (priv->focus == widget)
		return;

	if (priv->focus)
		g_object_unref (priv->focus);
	priv->focus = widget;
	if (priv->focus)
		g_object_ref (priv->focus);
}

void
nmt_newt_form_widget_activated (NmtNewtForm   *form,
                                NmtNewtWidget *widget,
                                newtComponent  co)
{
	NmtNewtFormPrivate *priv = NMT_NEWT_FORM_GET_PRIVATE (form);

	g_return_if_fail (priv->form != NULL);

	priv->activated = widget;
	priv->activated_co = co;
	newtFormSetTimer (priv->form, 1);
}

static void
nmt_newt_form_set_property (GObject      *object,
                            guint         prop_id,
                            const GValue *value,
                            GParamSpec   *pspec)
{
	NmtNewtFormPrivate *priv = NMT_NEWT_FORM_GET_PRIVATE (object);
	int screen_width, screen_height;

	switch (prop_id) {
	case PROP_TITLE:
		if (g_value_get_string (value)) {
			priv->title_lc = nmt_newt_locale_from_utf8 (g_value_get_string (value));
		} else
			priv->title_lc = NULL;
		break;
	case PROP_FULLSCREEN:
		if (g_value_get_boolean (value)) {
			newtGetScreenSize (&screen_width, &screen_height);
			priv->x = priv->y = 2;
			priv->fixed_x = priv->fixed_y = TRUE;
			priv->width = screen_width - 4;
			priv->height = screen_height - 4;
			priv->fixed_width = priv->fixed_height = TRUE;
		}
		break;
	case PROP_X:
		if (g_value_get_uint (value)) {
			priv->x = g_value_get_uint (value);
			priv->fixed_x = TRUE;
		}
		break;
	case PROP_Y:
		if (g_value_get_uint (value)) {
			priv->y = g_value_get_uint (value);
			priv->fixed_y = TRUE;
		}
		break;
	case PROP_WIDTH:
		if (g_value_get_uint (value)) {
			priv->width = g_value_get_uint (value);
			priv->fixed_width = TRUE;
		}
		break;
	case PROP_HEIGHT:
		if (g_value_get_uint (value)) {
			priv->height = g_value_get_uint (value);
			priv->fixed_height = TRUE;
		}
		break;
	case PROP_PADDING:
		priv->padding = g_value_get_uint (value);
		break;
	case PROP_ESCAPE_EXITS:
		priv->escape_exits = g_value_get_boolean (value);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

static void
nmt_newt_form_get_property (GObject    *object,
                            guint       prop_id,
                            GValue     *value,
                            GParamSpec *pspec)
{
	NmtNewtFormPrivate *priv = NMT_NEWT_FORM_GET_PRIVATE (object);

	switch (prop_id) {
	case PROP_TITLE:
		if (priv->title_lc) {
			g_value_take_string (value, nmt_newt_locale_to_utf8 (priv->title_lc));
		} else
			g_value_set_string (value, NULL);
		break;
	case PROP_X:
		g_value_set_uint (value, priv->x);
		break;
	case PROP_Y:
		g_value_set_uint (value, priv->y);
		break;
	case PROP_WIDTH:
		g_value_set_uint (value, priv->width);
		break;
	case PROP_HEIGHT:
		g_value_set_uint (value, priv->height);
		break;
	case PROP_PADDING:
		g_value_set_uint (value, priv->padding);
		break;
	case PROP_ESCAPE_EXITS:
		g_value_set_boolean (value, priv->escape_exits);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

static void
nmt_newt_form_class_init (NmtNewtFormClass *form_class)
{
	GObjectClass *object_class = G_OBJECT_CLASS (form_class);
	NmtNewtContainerClass *container_class = NMT_NEWT_CONTAINER_CLASS (form_class);
	NmtNewtWidgetClass *widget_class = NMT_NEWT_WIDGET_CLASS (form_class);

	g_type_class_add_private (form_class, sizeof (NmtNewtFormPrivate));

	/* virtual methods */
	object_class->set_property = nmt_newt_form_set_property;
	object_class->get_property = nmt_newt_form_get_property;
	object_class->finalize     = nmt_newt_form_finalize;

	widget_class->needs_rebuild = nmt_newt_form_needs_rebuild;

	container_class->remove = nmt_newt_form_remove;

	form_class->run = nmt_newt_form_real_run;

	g_object_class_install_property (object_class, PROP_TITLE,
	                                 g_param_spec_string ("title", "", "",
	                                                      NULL,
	                                                      G_PARAM_READWRITE |
	                                                      G_PARAM_STATIC_STRINGS |
	                                                      G_PARAM_CONSTRUCT_ONLY));
	g_object_class_install_property (object_class, PROP_FULLSCREEN,
	                                 g_param_spec_boolean ("fullscreen", "", "",
	                                                       FALSE,
	                                                       G_PARAM_WRITABLE |
	                                                       G_PARAM_STATIC_STRINGS |
	                                                       G_PARAM_CONSTRUCT_ONLY));
	g_object_class_install_property (object_class, PROP_X,
	                                 g_param_spec_uint ("x", "", "",
	                                                    0, G_MAXUINT, 0,
	                                                    G_PARAM_READWRITE |
	                                                    G_PARAM_STATIC_STRINGS |
	                                                    G_PARAM_CONSTRUCT_ONLY));
	g_object_class_install_property (object_class, PROP_Y,
	                                 g_param_spec_uint ("y", "", "",
	                                                    0, G_MAXUINT, 0,
	                                                    G_PARAM_READWRITE |
	                                                    G_PARAM_STATIC_STRINGS |
	                                                    G_PARAM_CONSTRUCT_ONLY));
	g_object_class_install_property (object_class, PROP_WIDTH,
	                                 g_param_spec_uint ("width", "", "",
	                                                    0, G_MAXUINT, 0,
	                                                    G_PARAM_READWRITE |
	                                                    G_PARAM_STATIC_STRINGS |
	                                                    G_PARAM_CONSTRUCT_ONLY));
	g_object_class_install_property (object_class, PROP_HEIGHT,
	                                 g_param_spec_uint ("height", "", "",
	                                                    0, G_MAXUINT, 0,
	                                                    G_PARAM_READWRITE |
	                                                    G_PARAM_STATIC_STRINGS |
	                                                    G_PARAM_CONSTRUCT_ONLY));
	g_object_class_install_property (object_class, PROP_PADDING,
	                                 g_param_spec_uint ("padding", "", "",
	                                                    0, G_MAXUINT, 1,
	                                                    G_PARAM_READWRITE |
	                                                    G_PARAM_STATIC_STRINGS |
	                                                    G_PARAM_CONSTRUCT_ONLY));
	g_object_class_install_property (object_class, PROP_ESCAPE_EXITS,
	                                 g_param_spec_boolean ("escape-exits", "", "",
	                                                       FALSE,
	                                                       G_PARAM_READWRITE |
	                                                       G_PARAM_STATIC_STRINGS |
	                                                       G_PARAM_CONSTRUCT_ONLY));
}
