/* -*- Mode: C; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*- */
/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright 2013 Red Hat, Inc.
 */

#include "config.h"

#include "nmt-newt-widget.h"
#include "nmt-newt-form.h"

G_DEFINE_ABSTRACT_TYPE (NmtNewtWidget, nmt_newt_widget, G_TYPE_INITIALLY_UNOWNED)

#define NMT_NEWT_WIDGET_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), NMT_TYPE_NEWT_WIDGET, NmtNewtWidgetPrivate))

typedef struct {
	NmtNewtWidget *parent;
	gboolean visible, realized, valid;
	gboolean exit_on_activate;
} NmtNewtWidgetPrivate;

enum {
	PROP_0,

	PROP_PARENT,
	PROP_VISIBLE,
	PROP_VALID,
	PROP_EXIT_ON_ACTIVATE,

	LAST_PROP
};

enum {
	NEEDS_REBUILD,

	LAST_SIGNAL
};

static guint signals[LAST_SIGNAL] = { 0 };

static void
nmt_newt_widget_init (NmtNewtWidget *widget)
{
	NmtNewtWidgetPrivate *priv = NMT_NEWT_WIDGET_GET_PRIVATE (widget);

	priv->visible = TRUE;
	priv->valid = TRUE;
}

static void
nmt_newt_widget_finalize (GObject *object)
{
	NmtNewtWidgetPrivate *priv = NMT_NEWT_WIDGET_GET_PRIVATE (object);

	nmt_newt_widget_unrealize (NMT_NEWT_WIDGET (object));
	g_clear_object (&priv->parent);

	G_OBJECT_CLASS (nmt_newt_widget_parent_class)->finalize (object);
}

void
nmt_newt_widget_realize (NmtNewtWidget *widget)
{
	NmtNewtWidgetPrivate *priv = NMT_NEWT_WIDGET_GET_PRIVATE (widget);

	if (!priv->realized) {
		NMT_NEWT_WIDGET_GET_CLASS (widget)->realize (widget);
		priv->realized = TRUE;
	}
}

void
nmt_newt_widget_unrealize (NmtNewtWidget *widget)
{
	NmtNewtWidgetPrivate *priv = NMT_NEWT_WIDGET_GET_PRIVATE (widget);

	if (priv->realized) {
		NMT_NEWT_WIDGET_GET_CLASS (widget)->unrealize (widget);
		priv->realized = FALSE;
	}
}

newtComponent *
nmt_newt_widget_get_components (NmtNewtWidget *widget)
{
	if (nmt_newt_widget_get_visible (widget)) {
		nmt_newt_widget_realize (widget);
		return NMT_NEWT_WIDGET_GET_CLASS (widget)->get_components (widget);
	} else
		return NULL;
}

void
nmt_newt_widget_size_request (NmtNewtWidget *widget,
                              int           *width,
                              int           *height)
{
	if (nmt_newt_widget_get_visible (widget)) {
		nmt_newt_widget_realize (widget);
		NMT_NEWT_WIDGET_GET_CLASS (widget)->size_request (widget, width, height);
	} else
		*width = *height = 0;
}

void
nmt_newt_widget_size_allocate (NmtNewtWidget *widget,
                               int            x,
                               int            y,
                               int            width,
                               int            height)
{
	if (nmt_newt_widget_get_visible (widget)) {
		nmt_newt_widget_realize (widget);
		NMT_NEWT_WIDGET_GET_CLASS (widget)->size_allocate (widget, x, y, width, height);
	}
}

newtComponent
nmt_newt_widget_get_focus_component (NmtNewtWidget *widget)
{
	if (!NMT_NEWT_WIDGET_GET_CLASS (widget)->get_focus_component)
		return NULL;

	return NMT_NEWT_WIDGET_GET_CLASS (widget)->get_focus_component (widget);
}

gboolean
nmt_newt_widget_activate (NmtNewtWidget *widget,
                          newtComponent  co)
{
	NmtNewtWidgetPrivate *priv = NMT_NEWT_WIDGET_GET_PRIVATE (widget);
	gboolean activated;

	if (!NMT_NEWT_WIDGET_GET_CLASS (widget)->activate)
		return FALSE;

	g_object_ref (widget);
	activated = NMT_NEWT_WIDGET_GET_CLASS (widget)->activate (widget, co);
	if (priv->exit_on_activate)
		nmt_newt_form_quit (nmt_newt_widget_get_form (widget));
	g_object_unref (widget);

	return activated;
}

gboolean
nmt_newt_widget_get_exit_on_activate (NmtNewtWidget *widget)
{
	NmtNewtWidgetPrivate *priv = NMT_NEWT_WIDGET_GET_PRIVATE (widget);

	return priv->exit_on_activate;
}

void
nmt_newt_widget_set_exit_on_activate (NmtNewtWidget *widget,
                                      gboolean       exit_on_activate)
{
	NmtNewtWidgetPrivate *priv = NMT_NEWT_WIDGET_GET_PRIVATE (widget);

	exit_on_activate = !!exit_on_activate;
	if (priv->exit_on_activate != exit_on_activate) {
		priv->exit_on_activate = exit_on_activate;
		g_object_notify (G_OBJECT (widget), "exit-on-activate");
	}
} 

gboolean
nmt_newt_widget_get_visible (NmtNewtWidget *widget)
{
	NmtNewtWidgetPrivate *priv = NMT_NEWT_WIDGET_GET_PRIVATE (widget);

	return priv->visible;
}

void
nmt_newt_widget_set_visible (NmtNewtWidget *widget,
                             gboolean       visible)
{
	NmtNewtWidgetPrivate *priv = NMT_NEWT_WIDGET_GET_PRIVATE (widget);

	visible = !!visible;
	if (priv->visible != visible) {
		priv->visible = visible;
		g_object_notify (G_OBJECT (widget), "visible");
		nmt_newt_widget_needs_rebuild (widget);
	}
}

void
nmt_newt_widget_set_parent (NmtNewtWidget *widget,
                            NmtNewtWidget *parent)
{
	NmtNewtWidgetPrivate *priv = NMT_NEWT_WIDGET_GET_PRIVATE (widget);

	g_clear_object (&priv->parent);
	priv->parent = parent ? g_object_ref (parent) : NULL;
	g_object_notify (G_OBJECT (widget), "parent");
}

NmtNewtWidget *
nmt_newt_widget_get_parent (NmtNewtWidget *widget)
{
	NmtNewtWidgetPrivate *priv = NMT_NEWT_WIDGET_GET_PRIVATE (widget);

	return priv->parent;
}

NmtNewtForm *
nmt_newt_widget_get_form (NmtNewtWidget *widget)
{
	while (widget) {
		if (NMT_IS_NEWT_FORM (widget))
			return NMT_NEWT_FORM (widget);
		widget = nmt_newt_widget_get_parent (widget);
	}

	return NULL;
}

gboolean
nmt_newt_widget_get_valid (NmtNewtWidget *widget)
{
	NmtNewtWidgetPrivate *priv = NMT_NEWT_WIDGET_GET_PRIVATE (widget);

	return priv->valid;
}

void
nmt_newt_widget_set_valid (NmtNewtWidget *widget,
                           gboolean       valid)
{
	NmtNewtWidgetPrivate *priv = NMT_NEWT_WIDGET_GET_PRIVATE (widget);

	valid = !!valid;
	if (priv->valid == valid)
		return;

	priv->valid = valid;
	g_object_notify (G_OBJECT (widget), "valid");
}

void
nmt_newt_widget_needs_rebuild (NmtNewtWidget *widget)
{
	g_signal_emit (widget, signals[NEEDS_REBUILD], 0);
}

static void
nmt_newt_widget_set_property (GObject      *object,
                              guint         prop_id,
                              const GValue *value,
                              GParamSpec   *pspec)
{
	NmtNewtWidget *widget = NMT_NEWT_WIDGET (object);

	switch (prop_id) {
	case PROP_PARENT:
		nmt_newt_widget_set_parent (widget, g_value_get_object (value));
		break;
	case PROP_VISIBLE:
		nmt_newt_widget_set_visible (widget, g_value_get_boolean (value));
		break;
	case PROP_EXIT_ON_ACTIVATE:
		nmt_newt_widget_set_exit_on_activate (widget, g_value_get_boolean (value));
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

static void
nmt_newt_widget_get_property (GObject    *object,
                              guint       prop_id,
                              GValue     *value,
                              GParamSpec *pspec)
{
	NmtNewtWidgetPrivate *priv = NMT_NEWT_WIDGET_GET_PRIVATE (object);

	switch (prop_id) {
	case PROP_PARENT:
		g_value_set_object (value, priv->parent);
		break;
	case PROP_VISIBLE:
		g_value_set_boolean (value, priv->visible);
		break;
	case PROP_VALID:
		g_value_set_boolean (value, priv->valid);
		break;
	case PROP_EXIT_ON_ACTIVATE:
		g_value_set_boolean (value, priv->exit_on_activate);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

static void
nmt_newt_widget_class_init (NmtNewtWidgetClass *widget_class)
{
	GObjectClass *object_class = G_OBJECT_CLASS (widget_class);

	g_type_class_add_private (widget_class, sizeof (NmtNewtWidgetPrivate));

	/* virtual methods */
	object_class->set_property = nmt_newt_widget_set_property;
	object_class->get_property = nmt_newt_widget_get_property;
	object_class->finalize     = nmt_newt_widget_finalize;

	/* signals */
	signals[NEEDS_REBUILD] =
		g_signal_new ("needs-rebuild",
		              G_OBJECT_CLASS_TYPE (object_class),
		              G_SIGNAL_RUN_LAST,
		              G_STRUCT_OFFSET (NmtNewtWidgetClass, needs_rebuild),
		              NULL, NULL, NULL,
		              G_TYPE_NONE, 0);

	/* properties */
	g_object_class_install_property (object_class, PROP_PARENT,
	                                 g_param_spec_object ("parent", "", "",
	                                                      NMT_TYPE_NEWT_WIDGET,
	                                                      G_PARAM_READABLE |
	                                                      G_PARAM_STATIC_STRINGS));
	g_object_class_install_property (object_class, PROP_VISIBLE,
	                                 g_param_spec_boolean ("visible", "", "",
	                                                       TRUE,
	                                                       G_PARAM_READWRITE |
	                                                       G_PARAM_STATIC_STRINGS));
	g_object_class_install_property (object_class, PROP_VALID,
	                                 g_param_spec_boolean ("valid", "", "",
	                                                       TRUE,
	                                                       G_PARAM_READABLE |
	                                                       G_PARAM_STATIC_STRINGS));
	g_object_class_install_property (object_class, PROP_EXIT_ON_ACTIVATE,
	                                 g_param_spec_boolean ("exit-on-activate", "", "",
	                                                       FALSE,
	                                                       G_PARAM_READWRITE |
	                                                       G_PARAM_STATIC_STRINGS));
}
