/* -*- Mode: C; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*- */
/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright 2013 Red Hat, Inc.
 */

#include "config.h"

#include <glib/gi18n-lib.h>

#include <nm-remote-connection.h>

#include "nmt-slave-list.h"
#include "nmt-connection-utils.h"

G_DEFINE_TYPE (NmtSlaveList, nmt_slave_list, NMT_TYPE_CONNECTION_LIST)

#define NMT_SLAVE_LIST_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), NMT_TYPE_SLAVE_LIST, NmtSlaveListPrivate))

typedef struct {
	NMConnection *master;
	const char *master_type, *master_uuid;
	int min_slaves;

	char *dialog_text;
	NmtAddConnectionTypeFilter type_filter;
	gpointer type_filter_data;
} NmtSlaveListPrivate;

enum {
	PROP_0,
	PROP_MASTER,
	PROP_MIN_SLAVES,
	PROP_DIALOG_TEXT,
	PROP_TYPE_FILTER,
	PROP_TYPE_FILTER_DATA,

	LAST_PROP
};

static gboolean nmt_slave_list_connection_filter (NmtConnectionList *list,
                                                  NMConnection      *connection,
                                                  gpointer           user_data);

NmtNewtWidget *
nmt_slave_list_new (NMConnection               *master,
                    int                         min_slaves,
                    const char                 *dialog_text,
                    NmtAddConnectionTypeFilter  type_filter,
                    gpointer                    type_filter_data)
{
	return g_object_new (NMT_TYPE_SLAVE_LIST,
	                     "master", master,
	                     "min-slaves", min_slaves,
	                     "dialog-text", dialog_text,
	                     "type-filter", type_filter,
	                     "type-filter-data", type_filter_data,
	                     "grouped", FALSE,
	                     "connection-filter", nmt_slave_list_connection_filter,
	                     NULL);
}

static void
nmt_slave_list_init (NmtSlaveList *list)
{
}

static void
nmt_slave_list_finalize (GObject *object)
{
	NmtSlaveListPrivate *priv = NMT_SLAVE_LIST_GET_PRIVATE (object);

	g_free (priv->dialog_text);
	g_object_unref (priv->master);

	G_OBJECT_CLASS (nmt_slave_list_parent_class)->finalize (object);
}

static gboolean
nmt_slave_list_connection_filter (NmtConnectionList *list,
                                  NMConnection      *connection,
                                  gpointer           user_data)
{
	NmtSlaveListPrivate *priv = NMT_SLAVE_LIST_GET_PRIVATE (list);
	NMSettingConnection *s_con;
	const char *master, *master_ifname, *slave_type;

	s_con = nm_connection_get_setting_connection (connection);
	g_return_val_if_fail (s_con != NULL, FALSE);

	slave_type = nm_setting_connection_get_slave_type (s_con);
	if (g_strcmp0 (slave_type, priv->master_type) != 0)
		return FALSE;

	master = nm_setting_connection_get_master (s_con);
	if (!master)
		return FALSE;

	master_ifname = nm_connection_get_virtual_iface_name (priv->master);
	if (g_strcmp0 (master, master_ifname) != 0 && g_strcmp0 (master, priv->master_uuid) != 0)
		return FALSE;

	return TRUE;
}

static void
nmt_slave_list_add_connection (NmtConnectionList *list)
{
	NmtSlaveListPrivate *priv = NMT_SLAVE_LIST_GET_PRIVATE (list);

	nmt_add_connection_full (priv->dialog_text, NULL,
	                         priv->type_filter, priv->type_filter_data);
}

static void
nmt_slave_list_edit_connection (NmtConnectionList *list,
                                NMConnection      *connection)
{
	nmt_edit_connection (connection);
}

static void
nmt_slave_list_remove_connection (NmtConnectionList  *list,
                                  NMRemoteConnection *connection)
{
	nmt_remove_connection (connection);
}

static void
nmt_slave_list_notify (GObject    *object,
                       GParamSpec *pspec)
{
	NmtSlaveListPrivate *priv = NMT_SLAVE_LIST_GET_PRIVATE (object);

	if (   priv->min_slaves
	    && !strcmp (pspec->name, "connections")) {
		int num;

		g_object_get (object, "num-connections", &num, NULL);
		nmt_newt_widget_set_valid (NMT_NEWT_WIDGET (object), num >= priv->min_slaves);
	}

	if (G_OBJECT_CLASS (nmt_slave_list_parent_class)->notify)
		G_OBJECT_CLASS (nmt_slave_list_parent_class)->notify (object, pspec);
}

static void
nmt_slave_list_set_property (GObject      *object,
                             guint         prop_id,
                             const GValue *value,
                             GParamSpec   *pspec)
{
	NmtSlaveListPrivate *priv = NMT_SLAVE_LIST_GET_PRIVATE (object);

	switch (prop_id) {
	case PROP_MASTER:
		priv->master = g_value_dup_object (value);
		if (priv->master) {
			NMSettingConnection *s_con = nm_connection_get_setting_connection (priv->master);

			priv->master_type = nm_setting_connection_get_connection_type (s_con);
			priv->master_uuid = nm_setting_connection_get_uuid (s_con);
		}
		break;
	case PROP_MIN_SLAVES:
		priv->min_slaves = g_value_get_int (value);
		break;
	case PROP_DIALOG_TEXT:
		priv->dialog_text = g_value_dup_string (value);
		break;
	case PROP_TYPE_FILTER:
		priv->type_filter = g_value_get_pointer (value);
		break;
	case PROP_TYPE_FILTER_DATA:
		priv->type_filter_data = g_value_get_pointer (value);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

static void
nmt_slave_list_get_property (GObject    *object,
                             guint       prop_id,
                             GValue     *value,
                             GParamSpec *pspec)
{
	NmtSlaveListPrivate *priv = NMT_SLAVE_LIST_GET_PRIVATE (object);

	switch (prop_id) {
	case PROP_MASTER:
		g_value_set_object (value, priv->master);
		break;
	case PROP_MIN_SLAVES:
		g_value_set_int (value, priv->min_slaves);
		break;
	case PROP_DIALOG_TEXT:
		g_value_set_string (value, priv->dialog_text);
		break;
	case PROP_TYPE_FILTER:
		g_value_set_pointer (value, priv->type_filter);
		break;
	case PROP_TYPE_FILTER_DATA:
		g_value_set_pointer (value, priv->type_filter_data);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

static void
nmt_slave_list_class_init (NmtSlaveListClass *list_class)
{
	GObjectClass *object_class = G_OBJECT_CLASS (list_class);
	NmtConnectionListClass *connection_list_class = NMT_CONNECTION_LIST_CLASS (list_class);

	g_type_class_add_private (list_class, sizeof (NmtSlaveListPrivate));

	/* virtual methods */
	object_class->notify       = nmt_slave_list_notify;
	object_class->set_property = nmt_slave_list_set_property;
	object_class->get_property = nmt_slave_list_get_property;
	object_class->finalize     = nmt_slave_list_finalize;

	connection_list_class->add_connection    = nmt_slave_list_add_connection;
	connection_list_class->edit_connection   = nmt_slave_list_edit_connection;
	connection_list_class->remove_connection = nmt_slave_list_remove_connection;

	g_object_class_install_property (object_class, PROP_MASTER,
	                                 g_param_spec_object ("master", "", "",
	                                                      NM_TYPE_CONNECTION,
	                                                      G_PARAM_READWRITE |
	                                                      G_PARAM_CONSTRUCT_ONLY |
	                                                      G_PARAM_STATIC_STRINGS));
	g_object_class_install_property (object_class, PROP_MIN_SLAVES,
	                                 g_param_spec_int ("min-slaves", "", "",
	                                                   0, G_MAXINT, 0,
	                                                   G_PARAM_READWRITE |
	                                                   G_PARAM_CONSTRUCT_ONLY |
	                                                   G_PARAM_STATIC_STRINGS));
	g_object_class_install_property (object_class, PROP_DIALOG_TEXT,
	                                 g_param_spec_string ("dialog-text", "", "",
	                                                      NULL,
	                                                      G_PARAM_READWRITE |
	                                                      G_PARAM_CONSTRUCT_ONLY |
	                                                      G_PARAM_STATIC_STRINGS));
	g_object_class_install_property (object_class, PROP_TYPE_FILTER,
	                                 g_param_spec_pointer ("type-filter", "", "",
	                                                       G_PARAM_READWRITE |
	                                                       G_PARAM_CONSTRUCT_ONLY |
	                                                       G_PARAM_STATIC_STRINGS));
	g_object_class_install_property (object_class, PROP_TYPE_FILTER_DATA,
	                                 g_param_spec_pointer ("type-filter-data", "", "",
	                                                       G_PARAM_READWRITE |
	                                                       G_PARAM_CONSTRUCT_ONLY |
	                                                       G_PARAM_STATIC_STRINGS));
}
